#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester
import unittest

from src.resiliens.retryable.Retryable import Retryable


class TestRetryable(unittest.TestCase):
    MAX_ATTEMPTS: int = 5
    BACKOFF: int = 100

    failed_count: int
    successful_count: int

    def setUp(self) -> None:
        self.failed_count = 0
        self.successful_count = 0

    @Retryable(max_retries=MAX_ATTEMPTS)
    def fake_successful_http_call(self):
        self.successful_count += 1
        return True

    @Retryable(max_retries=MAX_ATTEMPTS)
    def fake_failed_http_call(self):
        self.failed_count += 1
        raise Exception()

    def test_callRaisesException_retryIsDoneRightNumberOfTimes(self):
        self.fake_failed_http_call()
        expected = self.MAX_ATTEMPTS
        actual = self.failed_count

        self.assertEqual(expected, actual)

    def test_callSucceeds_noRetryIsPerformed(self):
        self.fake_successful_http_call()
        expected = 1
        actual = self.successful_count

        self.assertEqual(expected, actual)

    def test_callFailsWithUnexpectedException_noRetryIsPerformedAndExceptionIsRaised(self):
        @Retryable(max_retries=self.MAX_ATTEMPTS, expected_exception=IOError)
        def failed_http_call_with_unexpected_exception():
            self.failed_count += 1
            raise Exception()

        self.assertRaises(Exception, failed_http_call_with_unexpected_exception)



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

from typing import Any

from src.resiliens.circuit_breaker.CircuitBreakerStatus import CircuitBreakerStatus


class CircuitBreakerState:
    _status: str
    fail_count: int
    last_failure: Any
    opened: float

    def __init__(self, status: str,
                 fail_count: int = 0,
                 last_failure=None,
                 opened: float = 0):
        self.fail_count = fail_count
        self.status = status
        self.last_failure = last_failure
        self.opened = opened

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, new_status: str):
        if CircuitBreakerStatus.is_valid_status(new_status):
            self._status = new_status
        # Todo raise exception



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester
import time
from datetime import datetime
from functools import wraps
from inspect import isgeneratorfunction
from typing import Callable, Type, Any, Union


class Retryable:

    max_retries: int
    backoff: Union[int, float]
    backoff_exponent: Union[int, float]
    fallback_function: Callable

    _current_attempts: int
    _last_failure: Exception

    def __init__(self, max_retries: int = 3, backoff: Union[int, float] = 1000,
                 backoff_multiplier: Union[int, float] = None,
                 fallback: Callable = None,
                 expected_exception: Type[BaseException] = Exception):
        """

        :param max_retries: Max number of retries until it should give up.
        :param backoff: Backoff time in MILLISECONDS. If you don't set a backoff_exponent, this
        will be constant (e.g. 1000 milliseconds for all retries). If you do set a backoff_exponent,
        the backoff time will increase for every retry.
        :param backoff_exponent: Exponent to make the backoff time increase for every attempt
        (backoff * (retry_count ** backoff_exponent)). Say you set it to 2. On the second attempt,
        the backoff time will be multiplied with 2. On the fourth attempt, the backoff time will be multiplied with 8.
        :param fallback: Fallback function to get called in case the max retries limit has been reached. Optional,
        if not set the last exception will just get thrown.
        :param expected_exception: For what exceptions should we attempt to retry? Default is any exception, but you may
        want this to be more fine-grained (e.g. ConnectionError, RequestException)
        """
        self.max_retries = max_retries
        self.backoff = backoff / 1000  # Milliseconds to seconds
        self.fallback_function = fallback
        self._current_attempts = 0
        self.backoff_exponent = backoff_multiplier
        self._expected_exception = expected_exception

    def __call__(self, decorated_function):
        return self.decorate(decorated_function)

    def call(self, func, *args, **kwargs) -> Any:
        return func(*args, **kwargs)

    def call_generator(self, func, *args, **kwargs):
        for el in func(*args, **kwargs):
            yield el

    def get_backoff_time(self):
        if self.backoff_exponent:
            return max(self._current_attempts ** self.backoff_exponent, 1) * self.backoff
        else:
            return self.backoff

    def decorate(self, function_to_decorate) -> Callable:
        if isgeneratorfunction(function_to_decorate):
            call = self.call_generator
        else:
            call = self.call

        @wraps(function_to_decorate)
        def wrapper(*args, **kwargs):
            self.retry_if_needed(call, function_to_decorate, *args, **kwargs)

        return wrapper

    def retry_if_needed(self, call, function_to_decorate, *args, **kwargs):
        while self._current_attempts < self.max_retries:
            try:
                return call(function_to_decorate, *args, **kwargs)
            except Exception as e:
                if issubclass(e.__class__, self._expected_exception):
                    self._current_attempts += 1
                    self._last_failure = e
                    if self._current_attempts < self.max_retries:
                        time.sleep(self.get_backoff_time())
                else:
                    raise e



import unittest

from src.resiliens.circuit_breaker import CircuitBreaker
from src.resiliens.circuit_breaker import CircuitBreakerStatus


class TestCircuitBreaker(unittest.TestCase):
    circuit_breaker: CircuitBreaker

    def test_circuitBreakerIsGivenName_getNameProperty_returnsRightName(self):
        self.circuit_breaker = CircuitBreaker(name="CB1")

        expected = "CB1"
        actual = self.circuit_breaker.name

        self.assertEqual(expected, actual)

    def test_circuitBreakerIsGivenMaxAttempts_getMaxAttemptsProperty_returnsRightMaxAttempts(self):
        self.circuit_breaker = CircuitBreaker(max_attempts=100)

        expected = 100
        actual = self.circuit_breaker.max_attempts

        self.assertEqual(expected, actual)

    def test_circuitBreakerIsInitialized_circuitBreakerStatusIsClosed(self):
        self.circuit_breaker = CircuitBreaker()

        expected = CircuitBreakerStatus.CLOSED
        actual = self.circuit_breaker.status

        self.assertEqual(expected, actual)




from .manager import CircuitBreakerManager
from .CircuitBreaker import CircuitBreaker
from .CircuitBreakerException import CircuitBreakerException
from .CircuitBreakerStatus import CircuitBreakerStatus



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

from datetime import timedelta, datetime
from functools import wraps
from inspect import isgeneratorfunction
from math import ceil, floor
from time import monotonic
from types import TracebackType
from typing import Union, Callable, Optional, Type, Any

from src.resiliens.circuit_breaker.CircuitBreakerException import CircuitBreakerException
from src.resiliens.circuit_breaker.manager.CircuitBreakerManager import CircuitBreakerManager
from src.resiliens.circuit_breaker.CircuitBreakerState import CircuitBreakerState
from src.resiliens.circuit_breaker.CircuitBreakerStatus import CircuitBreakerStatus


class CircuitBreaker:

    _max_attempts: int
    _reset_timeout: Union[float, int]
    _expected_exception: Type[BaseException]
    _fallback_function: Callable
    _fallback_function_with_exception: Callable

    def __init__(self,
                 max_attempts: int = 5,
                 reset_timeout: Union[float, int] = 20_000,
                 expected_exception: Type[BaseException] = Exception,
                 name: str = None,
                 fallback_function: Callable = None,
                 fallback_function_with_exception: Callable = None):
        """

        :param max_attempts: Max failed attempts until the circuit breaker should be opened
        :param reset_timeout:
        Number of milliseconds until an opened circuit breaker should become half-opened and allow new attempts
        :param expected_exception: The
        exception the circuit breaker should expect as a failure (e.g. ConnectionError, RequestException)
        :param
        name: Name of the circuit breaker instance. Mostly useful if you intend to use the CircuitBreakerManager.
        :param fallback_function: A function to use as fallback if the circuit breaker is opened.
        :param fallback_function_with_exception: A function to use as fallback if the circuit breaker is opened.
        The first argument supplied to it will be the most recent exception (i.e.
        fallback_function_with_exception(last_exception, *args, **kwargs))
        """

        self._state = CircuitBreakerState(
            status=CircuitBreakerStatus.CLOSED,
            fail_count=0,
            last_failure=None,
            opened=monotonic()
        )
        self._max_attempts = max_attempts
        self._reset_timeout = reset_timeout / 1000  # From milliseconds to seconds
        self._expected_exception = expected_exception
        self._fallback_function = fallback_function
        self._fallback_function_with_exception = fallback_function_with_exception
        self._name = name

    @property
    def status(self):
        if self._state.status == CircuitBreakerStatus.OPEN and self.open_seconds_remaining <= 0:
            return CircuitBreakerStatus.HALF_OPEN
        return self._state.status

    @property
    def max_attempts(self):
        return self._max_attempts

    @property
    def open_until(self):
        return datetime.utcnow() + timedelta(seconds=self.open_seconds_remaining)

    @property
    def open_seconds_remaining(self) -> int:
        remain = (self._state.opened + self._reset_timeout) - monotonic()

        return ceil(remain) if remain > 0 else floor(remain)

    @property
    def failure_count(self):
        return self._state.fail_count

    @property
    def closed(self):
        return self.status == CircuitBreakerStatus.CLOSED

    @property
    def opened(self):
        return self.status == CircuitBreakerStatus.OPEN

    @property
    def name(self):
        return self._name

    @property
    def last_failure(self):
        return self._state.last_failure

    @property
    def fallback_function(self):
        if self._fallback_function is not None:
            return self._fallback_function
        return self._fallback_function_with_exception

    def __call__(self, decorated_function):
        return self.decorate(decorated_function)

    def __enter__(self):
        return None

    def __exit__(self,
                 exception_type: Optional[Type[BaseException]],
                 exception_value: Optional[BaseException],
                 exception_traceback: Optional[TracebackType]
                 ) -> bool:
        if exception_type and issubclass(exception_type, self._expected_exception):
            self._state.last_failure = exception_value
            self.__call_failed()
        else:
            self.__call_succeeded()
        return False

    def decorate(self, function_to_decorate) -> Callable:
        if self._name is None:
            self._name = function_to_decorate.__name__

        CircuitBreakerManager.register(self)

        if isgeneratorfunction(function_to_decorate):
            call = self.call_generator
        else:
            call = self.call

        @wraps(function_to_decorate)
        def wrapper(*args, **kwargs):
            if self.opened:
                if self.fallback_function:
                    return self.fallback_function(*args, **kwargs)
                elif self._fallback_function_with_exception:
                    return self._fallback_function_with_exception(self._state.last_failure, *args, **kwargs)
                raise CircuitBreakerException(self)
            return call(function_to_decorate, *args, **kwargs)

        return wrapper

    def call(self, func, *args, **kwargs) -> Any:
        with self:
            return func(*args, **kwargs)

    def call_generator(self, func, *args, **kwargs):
        with self:
            for el in func(*args, **kwargs):
                yield el

    def __call_succeeded(self) -> None:
        self._state.status = CircuitBreakerStatus.CLOSED
        self._state.last_failure = None
        self._state.fail_count = 0

    def __call_failed(self) -> None:
        self._state.fail_count += 1
        if self._state.fail_count >= self._max_attempts:
            self._state.status = CircuitBreakerStatus.OPEN
            self._state.opened = monotonic()

    def force_open(self) -> None:
        self._state.status = CircuitBreakerStatus.OPEN
        self._state.opened = monotonic()

    def force_reset(self) -> None:
        self.__call_succeeded()

    def __str__(self, *args, **kwargs) -> str:
        return self._name



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

from typing import Iterable


class CircuitBreakerManager(object):
    circuit_breakers = {}

    @classmethod
    def register(cls, circuit_breaker):
        cls.circuit_breakers[circuit_breaker.name] = circuit_breaker

    @classmethod
    def all_closed(cls) -> bool:
        return len(list(cls.get_open())) == 0

    @classmethod
    def get_circuits(cls):
        return cls.circuit_breakers.values()

    @classmethod
    def get(cls, name: str) -> "CircuitBreaker":
        return cls.circuit_breakers.get(name)

    @classmethod
    def get_open(cls) -> Iterable["CircuitBreaker"]:
        for circuit in cls.get_circuits():
            if circuit.opened:
                yield circuit

    @classmethod
    def get_closed(cls) -> Iterable["CircuitBreaker"]:
        for circuit in cls.get_circuits():
            if circuit.closed:
                yield circuit

    @classmethod
    def force_open(cls, name: str) -> None:
        circuit_breaker: "CircuitBreaker" = cls.circuit_breakers.get(name)
        circuit_breaker.force_open()

    @classmethod
    def force_reset(cls, name: str) -> None:
        circuit_breaker: "CircuitBreaker" = cls.circuit_breakers.get(name)
        circuit_breaker.force_reset()

    @classmethod
    def force_all_open(cls) -> None:
        for circuit_breaker in cls.circuit_breakers.values():
            circuit_breaker.force_open()

    @classmethod
    def force_all_reset(cls) -> None:
        for circuit_breaker in cls.circuit_breakers.values():
            circuit_breaker.force_reset()


#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

class CircuitBreakerException(Exception):
    def __init__(self, circuit_breaker, *args):
        super(CircuitBreakerException, self).__init__(*args)
        self._circuit_breaker = circuit_breaker

    def __str__(self, *args, **kwargs):
        return '[Circuit breaker: %s] Reached %d failures and will be open until' \
               ' %s (%d sec remaining) (_last_failure: %r)' % (
                   self._circuit_breaker.name,
                   self._circuit_breaker.failure_count,
                   self._circuit_breaker.open_until,
                   round(self._circuit_breaker.open_seconds_remaining),
                   self._circuit_breaker._last_failure,
               )



from time import sleep

from unittest.mock import Mock, patch
from pytest import raises

from src.resiliens.circuit_breaker import CircuitBreaker, CircuitBreakerManager, CircuitBreakerException, CircuitBreakerStatus


def fake_successful_http_call():
    return True


@CircuitBreaker()
def circuit_success():
    return fake_successful_http_call()


@CircuitBreaker(max_attempts=1, name="circuit_failure")
def circuit_failure():
    raise IOError()


@CircuitBreaker(max_attempts=1, name="circuit_generator_failure")
def circuit_generator_failure():
    fake_successful_http_call()
    yield 1
    raise IOError()


@CircuitBreaker(max_attempts=1, name="max_attempts_1")
def circuit_max_attempts_1():
    return fake_successful_http_call()


@CircuitBreaker(max_attempts=2, reset_timeout=1, name="max_attempts_2")
def circuit_max_attempts_2_timeout_1():
    return fake_successful_http_call()


@CircuitBreaker(max_attempts=3, reset_timeout=1, name="max_attempts_3")
def circuit_max_attempts_3_timeout_1():
    return fake_successful_http_call()


def test_circuit_pass_through():
    assert circuit_success() is True


def test_circuitbreaker_monitor():
    assert CircuitBreakerManager.all_closed() is True
    assert len(list(CircuitBreakerManager.get_circuits())) == 6
    assert len(list(CircuitBreakerManager.get_closed())) == 6
    assert len(list(CircuitBreakerManager.get_open())) == 0

    with raises(IOError):
        circuit_failure()

    assert CircuitBreakerManager.all_closed() is False
    assert len(list(CircuitBreakerManager.get_circuits())) == 6
    assert len(list(CircuitBreakerManager.get_closed())) == 5
    assert len(list(CircuitBreakerManager.get_open())) == 1


@patch('test_integration.fake_successful_http_call', return_value=True)
def test_threshold_hit_prevents_consequent_calls(mock_remote):
    # type: (Mock) -> None
    mock_remote.side_effect = IOError('Connection refused')
    circuitbreaker = CircuitBreakerManager.get('max_attempts_1')

    assert circuitbreaker.closed

    with raises(IOError):
        circuit_max_attempts_1()

    assert circuitbreaker.opened

    with raises(CircuitBreakerException):
        circuit_max_attempts_1()

    mock_remote.assert_called_once_with()


@patch('test_integration.fake_successful_http_call', return_value=True)
def test_circuitbreaker_recover_half_open(mock_remote):
    circuitbreaker = CircuitBreakerManager.get('max_attempts_3')

    assert circuitbreaker.closed
    assert circuitbreaker.status == CircuitBreakerStatus.CLOSED

    assert circuit_max_attempts_3_timeout_1()

    mock_remote.side_effect = IOError('Connection refused')

    with raises(IOError):
        circuit_max_attempts_3_timeout_1()
    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 1

    with raises(IOError):
        circuit_max_attempts_3_timeout_1()
    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 2

    with raises(IOError):
        circuit_max_attempts_3_timeout_1()

    assert circuitbreaker.opened
    assert circuitbreaker.status == CircuitBreakerStatus.OPEN
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    with raises(CircuitBreakerException):
        circuit_max_attempts_3_timeout_1()
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    with raises(CircuitBreakerException):
        circuit_max_attempts_3_timeout_1()
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    sleep(1)

    assert not circuitbreaker.closed
    assert circuitbreaker.open_seconds_remaining < 0
    assert circuitbreaker.status == CircuitBreakerStatus.HALF_OPEN

    with raises(IOError):
        circuit_max_attempts_3_timeout_1()
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 4
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    with raises(CircuitBreakerException):
        circuit_max_attempts_3_timeout_1()


@patch('test_integration.fake_successful_http_call', return_value=True)
def test_circuitbreaker_reopens_after_successful_calls(mock_remote):
    circuitbreaker = CircuitBreakerManager.get('max_attempts_2')

    assert str(circuitbreaker) == 'max_attempts_2'

    assert circuitbreaker.closed
    assert circuitbreaker.status == CircuitBreakerStatus.CLOSED
    assert circuitbreaker.failure_count == 0

    assert circuit_max_attempts_2_timeout_1()

    mock_remote.side_effect = IOError('Connection refused')

    with raises(IOError):
        circuit_max_attempts_2_timeout_1()
    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 1

    with raises(IOError):
        circuit_max_attempts_2_timeout_1()

    assert circuitbreaker.opened
    assert circuitbreaker.status == CircuitBreakerStatus.OPEN
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    with raises(CircuitBreakerException):
        circuit_max_attempts_2_timeout_1()
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    mock_remote.side_effect = None

    with raises(CircuitBreakerException):
        circuit_max_attempts_2_timeout_1()
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_seconds_remaining <= 1

    sleep(1)

    assert not circuitbreaker.closed
    assert circuitbreaker.failure_count == 2
    assert circuitbreaker.open_seconds_remaining < 0
    assert circuitbreaker.status == CircuitBreakerStatus.HALF_OPEN

    assert circuit_max_attempts_2_timeout_1()

    assert circuitbreaker.closed
    assert circuitbreaker.status == CircuitBreakerStatus.CLOSED
    assert circuitbreaker.failure_count == 0

    assert circuit_max_attempts_2_timeout_1()
    assert circuit_max_attempts_2_timeout_1()
    assert circuit_max_attempts_2_timeout_1()


@patch("test_integration.fake_successful_http_call", return_value=True)
def test_circuitbreaker_handles_generator_functions(mock_remote):
    circuitbreaker = CircuitBreakerManager.get("circuit_generator_failure")
    assert circuitbreaker.closed

    with raises(IOError):
        list(circuit_generator_failure())

    assert circuitbreaker.opened

    with raises(CircuitBreakerException):
        list(circuit_generator_failure())

    mock_remote.assert_called_once_with()



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

class _CircuitBreakerStatus:
    _CLOSED: str = 'CLOSED'
    _OPEN: str = 'OPEN'
    _HALF_OPEN: str = 'HALF_OPEN'

    @property
    def CLOSED(self) -> str:
        return self._CLOSED

    @property
    def OPEN(self) -> str:
        return self._OPEN

    @property
    def HALF_OPEN(self) -> str:
        return self._HALF_OPEN

    def is_valid_status(self, status: str):
        return status == self._OPEN or \
               status == self._HALF_OPEN or \
               status == self._CLOSED


CircuitBreakerStatus = _CircuitBreakerStatus()
